import json

from utils.exporters import TopologyExporter

SAMPLE_DATA = {
    'nodes': [
        {
            'id': '192.168.1.1',
            'label': 'gateway',
            'ip': '192.168.1.1',
            'type': 'gateway',
            'hostname': 'gateway',
            'state': 'up',
            'mac_address': 'AA:BB:CC:DD:EE:01',
            'vendor': 'Cisco',
            'open_ports': 0,
            'protocols': [],
        },
        {
            'id': '192.168.1.10',
            'label': 'web-server',
            'ip': '192.168.1.10',
            'type': 'server',
            'hostname': 'web-server',
            'state': 'up',
            'mac_address': 'AA:BB:CC:DD:EE:10',
            'vendor': 'Dell',
            'open_ports': 2,
            'protocols': ['tcp'],
        },
    ],
    'links': [
        {
            'source': '192.168.1.10',
            'target': '192.168.1.1',
            'weight': 1.0,
            'type': 'gateway',
        },
    ],
    'stats': {
        'total_nodes': 2,
        'total_links': 1,
        'node_types': {'gateway': 1, 'server': 1},
    },
}


class TestExportJSON:
    def test_valid_json(self):
        exporter = TopologyExporter()
        result = exporter.export_json(SAMPLE_DATA)
        parsed = json.loads(result)

        assert 'metadata' in parsed
        assert 'topology' in parsed

    def test_metadata_counts(self):
        exporter = TopologyExporter()
        result = exporter.export_json(SAMPLE_DATA)
        parsed = json.loads(result)

        assert parsed['metadata']['total_nodes'] == 2
        assert parsed['metadata']['total_links'] == 1

    def test_empty_data(self):
        exporter = TopologyExporter()
        result = exporter.export_json({'nodes': [], 'links': [], 'stats': {}})
        parsed = json.loads(result)

        assert parsed['metadata']['total_nodes'] == 0


class TestExportCSVNodes:
    def test_csv_header(self):
        exporter = TopologyExporter()
        result = exporter.export_csv_nodes(SAMPLE_DATA)
        lines = result.strip().splitlines()

        assert lines[0] == 'IP,Hostname,Type,MAC_Address,Vendor,Open_Ports,State'

    def test_csv_row_count(self):
        exporter = TopologyExporter()
        result = exporter.export_csv_nodes(SAMPLE_DATA)
        lines = result.strip().splitlines()

        assert len(lines) == 3  # header + 2 nodes

    def test_csv_content(self):
        exporter = TopologyExporter()
        result = exporter.export_csv_nodes(SAMPLE_DATA)

        assert '192.168.1.1' in result
        assert 'gateway' in result
        assert 'Cisco' in result


class TestExportCSVLinks:
    def test_csv_header(self):
        exporter = TopologyExporter()
        result = exporter.export_csv_links(SAMPLE_DATA)
        lines = result.strip().splitlines()

        assert lines[0] == 'Source,Target,Weight,Type'

    def test_csv_row_count(self):
        exporter = TopologyExporter()
        result = exporter.export_csv_links(SAMPLE_DATA)
        lines = result.strip().splitlines()

        assert len(lines) == 2  # header + 1 link

    def test_empty_links(self):
        exporter = TopologyExporter()
        result = exporter.export_csv_links({'links': []})
        lines = result.strip().splitlines()

        assert len(lines) == 1  # header only


class TestExportSVG:
    def test_produces_valid_svg(self):
        exporter = TopologyExporter()
        result = exporter.export_svg(SAMPLE_DATA)

        assert result.startswith('<svg')
        assert result.strip().endswith('</svg>')

    def test_contains_nodes(self):
        exporter = TopologyExporter()
        result = exporter.export_svg(SAMPLE_DATA)

        assert '192.168.1.1' in result
        assert '192.168.1.10' in result

    def test_contains_circles(self):
        exporter = TopologyExporter()
        result = exporter.export_svg(SAMPLE_DATA)

        assert '<circle' in result

    def test_contains_edges(self):
        exporter = TopologyExporter()
        result = exporter.export_svg(SAMPLE_DATA)

        assert '<line' in result

    def test_contains_legend(self):
        exporter = TopologyExporter()
        result = exporter.export_svg(SAMPLE_DATA)

        assert 'Gateway' in result
        assert 'Server' in result

    def test_empty_data(self):
        exporter = TopologyExporter()
        result = exporter.export_svg({'nodes': [], 'links': [], 'stats': {}})

        assert 'No hosts discovered' in result

    def test_escapes_special_chars(self):
        exporter = TopologyExporter()
        data = {
            'nodes': [{
                'id': '10.0.0.1',
                'label': '<script>alert(1)</script>',
                'ip': '10.0.0.1',
                'type': 'gateway',
            }],
            'links': [],
            'stats': {'total_nodes': 1, 'total_links': 0, 'node_types': {'gateway': 1}},
        }
        result = exporter.export_svg(data)

        assert '<script>' not in result
        assert '&lt;script&gt;' in result


class TestExportRaster:
    def test_png_magic_bytes(self):
        exporter = TopologyExporter()
        data = exporter._render_raster(SAMPLE_DATA, 400, 300, fmt='png')

        assert data[:4] == b'\x89PNG'

    def test_jpeg_magic_bytes(self):
        exporter = TopologyExporter()
        data = exporter._render_raster(SAMPLE_DATA, 400, 300, fmt='jpg')

        assert data[:2] == b'\xff\xd8'

    def test_png_nonempty(self):
        exporter = TopologyExporter()
        data = exporter._render_raster(SAMPLE_DATA, 400, 300, fmt='png')

        assert len(data) > 100

    def test_empty_data_png(self):
        exporter = TopologyExporter()
        empty = {'nodes': [], 'links': [], 'stats': {}}
        data = exporter._render_raster(empty, 400, 300, fmt='png')

        assert data[:4] == b'\x89PNG'


class TestExportImage:
    def test_svg_by_extension(self, tmp_path):
        exporter = TopologyExporter()
        path = str(tmp_path / 'out.svg')
        exporter.export_image(SAMPLE_DATA, path)

        with open(path, 'r') as f:
            content = f.read()
        assert content.startswith('<svg')

    def test_png_by_extension(self, tmp_path):
        exporter = TopologyExporter()
        path = str(tmp_path / 'out.png')
        exporter.export_image(SAMPLE_DATA, path)

        with open(path, 'rb') as f:
            assert f.read(4) == b'\x89PNG'

    def test_jpg_by_extension(self, tmp_path):
        exporter = TopologyExporter()
        path = str(tmp_path / 'out.jpg')
        exporter.export_image(SAMPLE_DATA, path)

        with open(path, 'rb') as f:
            assert f.read(2) == b'\xff\xd8'

    def test_jpeg_by_extension(self, tmp_path):
        exporter = TopologyExporter()
        path = str(tmp_path / 'out.jpeg')
        exporter.export_image(SAMPLE_DATA, path)

        with open(path, 'rb') as f:
            assert f.read(2) == b'\xff\xd8'

    def test_unsupported_extension(self):
        exporter = TopologyExporter()
        try:
            exporter.export_image(SAMPLE_DATA, 'out.bmp')
            assert False, "Should have raised ValueError"
        except ValueError as e:
            assert 'Unsupported' in str(e)


class TestGetExportFormats:
    def test_formats_list(self):
        exporter = TopologyExporter()
        formats = exporter.get_export_formats()

        assert 'json' in formats
        assert 'png' in formats
        assert 'jpg' in formats
        assert 'svg' in formats
        assert 'csv_nodes' in formats
        assert 'csv_links' in formats



import nmap
import ipaddress
import logging
from typing import List, Dict, Any
import time

class NetworkScanner:
    def __init__(self):
        self.nm = nmap.PortScanner()
        self.logger = logging.getLogger(__name__)
    
    def validate_subnet(self, subnet: str) -> bool:
        """Validate if the provided subnet is valid."""
        try:
            ipaddress.ip_network(subnet, strict=False)
            return True
        except ValueError:
            return False
    
    def scan_subnet(self, subnet: str, scan_type: str = 'tcp') -> Dict[str, Any]:
        """
        Scan a subnet for live hosts using Nmap.
        
        Args:
            subnet: Network subnet in CIDR notation (e.g., '192.168.1.0/24')
            scan_type: Type of scan ('ping', 'tcp', 'syn')
        
        Returns:
            Dictionary containing scan results
        """
        if not self.validate_subnet(subnet):
            raise ValueError(f"Invalid subnet format: {subnet}")
        
        scan_results = {
            'subnet': subnet,
            'hosts': [],
            'scan_time': time.time(),
            'total_hosts': 0
        }
        
        try:
            # Configure scan arguments based on type
            if scan_type == 'ping':
                arguments = '-sn'  # Ping scan only
            elif scan_type == 'tcp':
                arguments = '-sT -F'  # TCP connect scan with fast mode
            elif scan_type == 'syn':
                arguments = '-sS -F'  # SYN stealth scan with fast mode
            else:
                arguments = '-sn'
            
            self.logger.info(f"Starting {scan_type} scan of {subnet}")
            scan_result = self.nm.scan(hosts=subnet, arguments=arguments)
            
            for host in self.nm.all_hosts():
                host_info = {
                    'ip': host,
                    'hostname': '',
                    'state': self.nm[host].state(),
                    'protocols': [],
                    'open_ports': [],
                    'mac_address': '',
                    'vendor': ''
                }
                
                # Get hostname
                if 'hostnames' in self.nm[host] and self.nm[host]['hostnames']:
                    host_info['hostname'] = self.nm[host]['hostnames'][0]['name']
                
                # Get MAC address and vendor info
                if 'addresses' in self.nm[host]:
                    if 'mac' in self.nm[host]['addresses']:
                        host_info['mac_address'] = self.nm[host]['addresses']['mac']
                        if 'vendor' in self.nm[host]:
                            host_info['vendor'] = list(self.nm[host]['vendor'].values())[0] if self.nm[host]['vendor'] else ''
                
                # Get port information if available
                for protocol in self.nm[host].all_protocols():
                    host_info['protocols'].append(protocol)
                    ports = self.nm[host][protocol].keys()
                    for port in ports:
                        port_state = self.nm[host][protocol][port]['state']
                        if port_state == 'open':
                            host_info['open_ports'].append({
                                'port': port,
                                'protocol': protocol,
                                'service': self.nm[host][protocol][port].get('name', 'unknown')
                            })
                
                scan_results['hosts'].append(host_info)
            
            scan_results['total_hosts'] = len(scan_results['hosts'])
            self.logger.info(f"Scan completed. Found {scan_results['total_hosts']} hosts")
            
        except Exception as e:
            self.logger.error(f"Scan failed: {str(e)}")
            raise Exception(f"Network scan failed: {str(e)}")
        
        return scan_results
    
    def get_network_info(self, subnet: str) -> Dict[str, Any]:
        """Get basic network information."""
        try:
            network = ipaddress.ip_network(subnet, strict=False)
            return {
                'network_address': str(network.network_address),
                'broadcast_address': str(network.broadcast_address),
                'netmask': str(network.netmask),
                'total_addresses': network.num_addresses,
                'usable_addresses': network.num_addresses - 2 if network.num_addresses > 2 else 0,
                'prefix_length': network.prefixlen
            }
        except ValueError as e:
            raise ValueError(f"Invalid network: {str(e)}")


import json
import csv
import io
import math
from typing import Dict, Any, List, Tuple
from xml.sax.saxutils import escape

from PIL import Image, ImageDraw, ImageFont

from config import NODE_COLORS

NODE_RADII = {
    'gateway': 28,
    'server': 22,
    'printer': 18,
    'workstation': 16,
    'unknown': 14,
}

BG_COLOR = '#f8f9fa'
EDGE_COLOR = '#cccccc'
LABEL_COLOR = '#333333'
SUBLABEL_COLOR = '#888888'


def _hex_to_rgb(hex_color: str) -> Tuple[int, int, int]:
    h = hex_color.lstrip('#')
    return (int(h[0:2], 16), int(h[2:4], 16), int(h[4:6], 16))


class TopologyExporter:

    def export_json(self, topology_data: Dict[str, Any]) -> str:
        """Export topology data as JSON string."""
        export_data = {
            'metadata': {
                'export_format': 'network_topology_json',
                'version': '1.0',
                'total_nodes': len(topology_data.get('nodes', [])),
                'total_links': len(topology_data.get('links', [])),
                'node_types': topology_data.get('stats', {}).get('node_types', {}),
            },
            'topology': topology_data,
        }
        return json.dumps(export_data, indent=2, default=str)

    def export_csv_nodes(self, topology_data: Dict[str, Any]) -> str:
        """Export node data as CSV."""
        buf = io.StringIO()
        writer = csv.writer(buf)
        writer.writerow(['IP', 'Hostname', 'Type', 'MAC_Address', 'Vendor', 'Open_Ports', 'State'])

        for node in topology_data.get('nodes', []):
            writer.writerow([
                node.get('ip', ''),
                node.get('hostname', ''),
                node.get('type', ''),
                node.get('mac_address', ''),
                node.get('vendor', ''),
                node.get('open_ports', 0),
                node.get('state', ''),
            ])

        buf.seek(0)
        return buf.getvalue()

    def export_csv_links(self, topology_data: Dict[str, Any]) -> str:
        """Export link data as CSV."""
        buf = io.StringIO()
        writer = csv.writer(buf)
        writer.writerow(['Source', 'Target', 'Weight', 'Type'])

        for link in topology_data.get('links', []):
            writer.writerow([
                link.get('source', ''),
                link.get('target', ''),
                link.get('weight', 0),
                link.get('type', ''),
            ])

        buf.seek(0)
        return buf.getvalue()

    # ------------------------------------------------------------------ #
    #  Image export (PNG / JPEG / SVG)
    # ------------------------------------------------------------------ #

    def export_image(self, topology_data: Dict[str, Any], path: str,
                     width: int = 900, height: int = 700) -> None:
        """
        Export topology image. Format is detected from file extension.

        Supported: .png, .jpg/.jpeg, .svg
        """
        ext = path.rsplit('.', 1)[-1].lower() if '.' in path else ''

        if ext == 'svg':
            svg = self.export_svg(topology_data, width, height)
            with open(path, 'w', encoding='utf-8') as f:
                f.write(svg)
        elif ext in ('png', 'jpg', 'jpeg'):
            img_bytes = self._render_raster(topology_data, width, height, fmt=ext)
            with open(path, 'wb') as f:
                f.write(img_bytes)
        else:
            raise ValueError(
                f"Unsupported image format '.{ext}'. Use .png, .jpg, or .svg"
            )

    # ------------------------------------------------------------------ #
    #  Raster rendering (PNG / JPEG) via Pillow
    # ------------------------------------------------------------------ #

    def _render_raster(self, topology_data: Dict[str, Any],
                       width: int, height: int, fmt: str = 'png') -> bytes:
        """Render topology to PNG or JPEG bytes using Pillow."""
        nodes = topology_data.get('nodes', [])
        links = topology_data.get('links', [])

        img = Image.new('RGB', (width, height), _hex_to_rgb(BG_COLOR))
        draw = ImageDraw.Draw(img)

        try:
            font = ImageFont.truetype("arial.ttf", 12)
            font_small = ImageFont.truetype("arial.ttf", 10)
            font_title = ImageFont.truetype("arial.ttf", 18)
        except (OSError, IOError):
            font = ImageFont.load_default()
            font_small = font
            font_title = font

        if not nodes:
            draw.text(
                (width // 2, height // 2),
                "No hosts discovered",
                fill=_hex_to_rgb(SUBLABEL_COLOR),
                font=font_title,
                anchor='mm',
            )
            return self._image_to_bytes(img, fmt)

        positions = self._compute_layout(nodes, width, height)

        # Title
        stats = topology_data.get('stats', {})
        title = f"Network Topology — {stats.get('total_nodes', 0)} hosts"
        draw.text((width // 2, 25), title, fill=_hex_to_rgb(LABEL_COLOR),
                  font=font_title, anchor='mm')

        # Edges
        pos_map = {n['id']: positions[i] for i, n in enumerate(nodes)}
        edge_rgb = _hex_to_rgb(EDGE_COLOR)
        for link in links:
            src = pos_map.get(link.get('source'))
            tgt = pos_map.get(link.get('target'))
            if src and tgt:
                draw.line([src, tgt], fill=edge_rgb, width=2)

        # Nodes
        for i, node in enumerate(nodes):
            x, y = positions[i]
            ntype = node.get('type', 'unknown')
            color = _hex_to_rgb(NODE_COLORS.get(ntype, NODE_COLORS['unknown']))
            r = NODE_RADII.get(ntype, NODE_RADII['unknown'])

            draw.ellipse(
                [x - r, y - r, x + r, y + r],
                fill=color, outline=(255, 255, 255), width=2,
            )

            label = node.get('label') or node.get('ip', '')
            ip = node.get('ip', '')

            draw.text((x, y - r - 8), label, fill=_hex_to_rgb(LABEL_COLOR),
                      font=font, anchor='mb')
            if label != ip:
                draw.text((x, y - r - 22), ip, fill=_hex_to_rgb(SUBLABEL_COLOR),
                          font=font_small, anchor='mb')

        # Legend
        self._draw_legend(draw, width, font)

        return self._image_to_bytes(img, fmt)

    def _draw_legend(self, draw: ImageDraw.ImageDraw, width: int, font) -> None:
        items = [
            ('Gateway', NODE_COLORS['gateway']),
            ('Server', NODE_COLORS['server']),
            ('Workstation', NODE_COLORS['workstation']),
            ('Printer', NODE_COLORS['printer']),
        ]
        lx = width - 135
        ly = 50
        box_h = len(items) * 22 + 12

        draw.rounded_rectangle(
            [lx - 10, ly - 10, lx + 120, ly - 10 + box_h],
            radius=5, fill=(255, 255, 255), outline=(221, 221, 221),
        )

        for i, (label, color) in enumerate(items):
            cy = ly + i * 22
            r = 6
            draw.ellipse(
                [lx, cy - r, lx + 2 * r, cy + r],
                fill=_hex_to_rgb(color),
            )
            draw.text((lx + 18, cy), label, fill=(85, 85, 85), font=font, anchor='lm')

    @staticmethod
    def _image_to_bytes(img: Image.Image, fmt: str) -> bytes:
        buf = io.BytesIO()
        save_fmt = 'JPEG' if fmt in ('jpg', 'jpeg') else 'PNG'
        if save_fmt == 'JPEG':
            img = img.convert('RGB')
        img.save(buf, format=save_fmt, quality=95)
        buf.seek(0)
        return buf.getvalue()

    # ------------------------------------------------------------------ #
    #  SVG rendering (no Pillow needed)
    # ------------------------------------------------------------------ #

    def export_svg(self, topology_data: Dict[str, Any],
                   width: int = 900, height: int = 700) -> str:
        """Export topology as SVG string."""
        nodes = topology_data.get('nodes', [])
        links = topology_data.get('links', [])

        if not nodes:
            return self._empty_svg(width, height)

        positions = self._compute_layout(nodes, width, height)

        parts: List[str] = []
        parts.append(
            f'<svg xmlns="http://www.w3.org/2000/svg" '
            f'width="{width}" height="{height}" '
            f'viewBox="0 0 {width} {height}" '
            f'style="font-family:sans-serif;background:{BG_COLOR}">'
        )
        parts.append(self._svg_defs())

        stats = topology_data.get('stats', {})
        title = f"{stats.get('total_nodes', 0)} hosts"
        parts.append(
            f'<text x="{width // 2}" y="30" text-anchor="middle" '
            f'font-size="18" font-weight="bold" fill="{LABEL_COLOR}">'
            f'Network Topology — {escape(title)}</text>'
        )

        pos_map = {n['id']: positions[i] for i, n in enumerate(nodes)}
        for link in links:
            src = pos_map.get(link.get('source'))
            tgt = pos_map.get(link.get('target'))
            if src and tgt:
                parts.append(
                    f'<line x1="{src[0]}" y1="{src[1]}" '
                    f'x2="{tgt[0]}" y2="{tgt[1]}" '
                    f'stroke="{EDGE_COLOR}" stroke-width="1.5" />'
                )

        for i, node in enumerate(nodes):
            x, y = positions[i]
            ntype = node.get('type', 'unknown')
            color = NODE_COLORS.get(ntype, NODE_COLORS['unknown'])
            r = NODE_RADII.get(ntype, NODE_RADII['unknown'])
            label = escape(node.get('label') or node.get('ip', ''))
            ip = escape(node.get('ip', ''))

            parts.append(
                f'<circle cx="{x}" cy="{y}" r="{r}" '
                f'fill="{color}" stroke="#fff" stroke-width="2" />'
            )
            parts.append(
                f'<text x="{x}" y="{y - r - 6}" text-anchor="middle" '
                f'font-size="11" fill="{LABEL_COLOR}">{label}</text>'
            )
            if label != ip:
                parts.append(
                    f'<text x="{x}" y="{y - r - 18}" text-anchor="middle" '
                    f'font-size="9" fill="{SUBLABEL_COLOR}">{ip}</text>'
                )

        parts.append(self._svg_legend(width))
        parts.append('</svg>')
        return '\n'.join(parts)

    # ------------------------------------------------------------------ #
    #  Shared layout
    # ------------------------------------------------------------------ #

    def _compute_layout(self, nodes: List[Dict], width: int, height: int):
        """Star layout: gateway at center, others in a circle."""
        cx, cy = width // 2, height // 2
        radius = min(width, height) * 0.35

        gateways = [i for i, n in enumerate(nodes) if n.get('type') == 'gateway']
        others = [i for i in range(len(nodes)) if i not in gateways]

        positions = [(0.0, 0.0)] * len(nodes)

        if gateways:
            for gi in gateways:
                positions[gi] = (float(cx), float(cy))

        ring_nodes = others if gateways else list(range(len(nodes)))
        n_ring = len(ring_nodes)
        if n_ring == 0:
            return positions

        if not gateways and n_ring == 1:
            positions[ring_nodes[0]] = (float(cx), float(cy))
            return positions

        for idx, ni in enumerate(ring_nodes):
            angle = 2 * math.pi * idx / n_ring - math.pi / 2
            x = cx + radius * math.cos(angle)
            y = cy + radius * math.sin(angle)
            positions[ni] = (round(x, 1), round(y, 1))

        return positions

    # ------------------------------------------------------------------ #
    #  SVG helpers
    # ------------------------------------------------------------------ #

    def _svg_defs(self) -> str:
        return (
            '<defs><filter id="shadow" x="-20%" y="-20%" width="140%" height="140%">'
            '<feDropShadow dx="1" dy="1" stdDeviation="2" flood-opacity="0.15" />'
            '</filter></defs>'
        )

    def _svg_legend(self, width: int) -> str:
        items = [
            ('Gateway', NODE_COLORS['gateway']),
            ('Server', NODE_COLORS['server']),
            ('Workstation', NODE_COLORS['workstation']),
            ('Printer', NODE_COLORS['printer']),
        ]
        parts = []
        lx = width - 130
        ly = 55
        parts.append(
            f'<rect x="{lx - 10}" y="{ly - 15}" width="130" height="{len(items) * 22 + 10}" '
            f'rx="5" fill="white" stroke="#ddd" />'
        )
        for i, (label, color) in enumerate(items):
            y = ly + i * 22
            parts.append(f'<circle cx="{lx + 6}" cy="{y}" r="6" fill="{color}" />')
            parts.append(
                f'<text x="{lx + 18}" y="{y + 4}" font-size="11" fill="#555">{label}</text>'
            )
        return '\n'.join(parts)

    def _empty_svg(self, width: int, height: int) -> str:
        return (
            f'<svg xmlns="http://www.w3.org/2000/svg" '
            f'width="{width}" height="{height}" style="background:{BG_COLOR}">'
            f'<text x="{width // 2}" y="{height // 2}" text-anchor="middle" '
            f'font-size="16" fill="#999">No hosts discovered</text></svg>'
        )

    def get_export_formats(self) -> List[str]:
        """Get list of available export formats."""
        return ['json', 'png', 'jpg', 'svg', 'csv_nodes', 'csv_links']



"""whoson -- lightweight subnet audit tool."""

import argparse
import sys
import time

from discovery.nmap_scan import NetworkScanner
from graph.topology_builder import TopologyBuilder
from utils.exporters import TopologyExporter


def _print_table(scan_results):
    """Print a host table to stdout."""
    hosts = scan_results.get('hosts', [])
    if not hosts:
        print("No hosts found.")
        return

    builder = TopologyBuilder()

    rows = []
    for h in hosts:
        host_type = builder._classify_host(h)
        ports = ','.join(str(p['port']) for p in h.get('open_ports', []))
        rows.append({
            'ip': h.get('ip', ''),
            'hostname': h.get('hostname', '') or '-',
            'type': host_type,
            'mac': h.get('mac_address', '') or '-',
            'vendor': h.get('vendor', '') or '-',
            'ports': ports or '-',
        })

    col_widths = {
        'ip': max(len('IP'), max(len(r['ip']) for r in rows)),
        'hostname': max(len('Hostname'), max(len(r['hostname']) for r in rows)),
        'type': max(len('Type'), max(len(r['type']) for r in rows)),
        'mac': max(len('MAC'), max(len(r['mac']) for r in rows)),
        'vendor': max(len('Vendor'), max(len(r['vendor']) for r in rows)),
        'ports': max(len('Ports'), max(len(r['ports']) for r in rows)),
    }

    header = (
        f"{'IP':<{col_widths['ip']}}  "
        f"{'Hostname':<{col_widths['hostname']}}  "
        f"{'Type':<{col_widths['type']}}  "
        f"{'MAC':<{col_widths['mac']}}  "
        f"{'Vendor':<{col_widths['vendor']}}  "
        f"{'Ports':<{col_widths['ports']}}"
    )
    sep = '-' * len(header)

    print(header)
    print(sep)
    for r in rows:
        print(
            f"{r['ip']:<{col_widths['ip']}}  "
            f"{r['hostname']:<{col_widths['hostname']}}  "
            f"{r['type']:<{col_widths['type']}}  "
            f"{r['mac']:<{col_widths['mac']}}  "
            f"{r['vendor']:<{col_widths['vendor']}}  "
            f"{r['ports']:<{col_widths['ports']}}"
        )


def main(argv=None):
    parser = argparse.ArgumentParser(
        prog='whoson',
        description='Lightweight subnet audit tool -- see what is alive on your network.',
    )
    parser.add_argument(
        'subnet',
        help='Network in CIDR notation (e.g. 192.168.1.0/24)',
    )
    parser.add_argument(
        '-t', '--type',
        choices=['ping', 'tcp', 'syn'],
        default='tcp',
        help='Scan type (default: tcp)',
    )
    parser.add_argument(
        '-i', '--image',
        metavar='FILE',
        help='Save topology image (format from extension: .png, .jpg, .svg)',
    )
    parser.add_argument(
        '--json',
        metavar='FILE',
        dest='json_file',
        help='Save topology data as JSON',
    )
    parser.add_argument(
        '--csv',
        metavar='FILE',
        dest='csv_file',
        help='Save host list as CSV',
    )
    parser.add_argument(
        '-q', '--quiet',
        action='store_true',
        help='Suppress table output (only write files)',
    )

    args = parser.parse_args(argv)

    scanner = NetworkScanner()

    if not scanner.validate_subnet(args.subnet):
        print(f"Error: invalid subnet '{args.subnet}'", file=sys.stderr)
        sys.exit(1)

    info = scanner.get_network_info(args.subnet)
    if not args.quiet:
        print(f"Scanning {args.subnet}  ({info['usable_addresses']} usable addresses, {args.type} scan)")

    t0 = time.time()
    scan_results = scanner.scan_subnet(args.subnet, args.type)
    elapsed = time.time() - t0

    total = scan_results['total_hosts']
    if not args.quiet:
        print(f"Found {total} host{'s' if total != 1 else ''} in {elapsed:.1f}s\n")

    if total == 0:
        return

    if not args.quiet:
        _print_table(scan_results)

    builder = TopologyBuilder()
    topology_data = builder.build_topology(scan_results)
    exporter = TopologyExporter()

    if args.image:
        exporter.export_image(topology_data, args.image)
        if not args.quiet:
            print(f"\nImage saved to {args.image}")

    if args.json_file:
        json_str = exporter.export_json(topology_data)
        with open(args.json_file, 'w', encoding='utf-8') as f:
            f.write(json_str)
        if not args.quiet:
            print(f"JSON saved to {args.json_file}")

    if args.csv_file:
        csv_str = exporter.export_csv_nodes(topology_data)
        with open(args.csv_file, 'w', encoding='utf-8') as f:
            f.write(csv_str)
        if not args.quiet:
            print(f"CSV saved to {args.csv_file}")


if __name__ == '__main__':
    main()



from graph.topology_builder import TopologyBuilder


SAMPLE_SCAN = {
    'subnet': '192.168.1.0/24',
    'hosts': [
        {
            'ip': '192.168.1.1',
            'hostname': 'gateway',
            'state': 'up',
            'mac_address': 'AA:BB:CC:DD:EE:01',
            'vendor': 'Cisco',
            'open_ports': [],
            'protocols': [],
        },
        {
            'ip': '192.168.1.10',
            'hostname': 'web-server',
            'state': 'up',
            'mac_address': 'AA:BB:CC:DD:EE:10',
            'vendor': 'Dell',
            'open_ports': [
                {'port': 80, 'protocol': 'tcp', 'service': 'http'},
                {'port': 443, 'protocol': 'tcp', 'service': 'https'},
            ],
            'protocols': ['tcp'],
        },
        {
            'ip': '192.168.1.50',
            'hostname': '',
            'state': 'up',
            'mac_address': 'AA:BB:CC:DD:EE:50',
            'vendor': '',
            'open_ports': [],
            'protocols': [],
        },
        {
            'ip': '192.168.1.99',
            'hostname': 'office-printer',
            'state': 'up',
            'mac_address': 'AA:BB:CC:DD:EE:99',
            'vendor': 'HP',
            'open_ports': [
                {'port': 9100, 'protocol': 'tcp', 'service': 'jetdirect'},
            ],
            'protocols': ['tcp'],
        },
    ],
}


class TestClassifyHost:
    def setup_method(self):
        self.builder = TopologyBuilder()

    def test_gateway_dot_1(self):
        host = {'ip': '192.168.1.1', 'open_ports': []}
        assert self.builder._classify_host(host) == 'gateway'

    def test_gateway_dot_254(self):
        host = {'ip': '10.0.0.254', 'open_ports': []}
        assert self.builder._classify_host(host) == 'gateway'

    def test_server_with_http(self):
        host = {'ip': '192.168.1.10', 'open_ports': [{'port': 80}]}
        assert self.builder._classify_host(host) == 'server'

    def test_server_with_ssh(self):
        host = {'ip': '192.168.1.10', 'open_ports': [{'port': 22}]}
        assert self.builder._classify_host(host) == 'server'

    def test_printer(self):
        host = {'ip': '192.168.1.99', 'open_ports': [{'port': 9100}]}
        assert self.builder._classify_host(host) == 'printer'

    def test_workstation_default(self):
        host = {'ip': '192.168.1.50', 'open_ports': []}
        assert self.builder._classify_host(host) == 'workstation'

    def test_gateway_takes_priority_over_server(self):
        host = {'ip': '192.168.1.1', 'open_ports': [{'port': 80}]}
        assert self.builder._classify_host(host) == 'gateway'


class TestBuildTopology:
    def test_basic_build(self):
        builder = TopologyBuilder()
        result = builder.build_topology(SAMPLE_SCAN)

        assert result['stats']['total_nodes'] == 4
        assert len(result['nodes']) == 4
        assert len(result['links']) == 3

    def test_all_edges_connect_to_gateway(self):
        builder = TopologyBuilder()
        result = builder.build_topology(SAMPLE_SCAN)

        for link in result['links']:
            assert '192.168.1.1' in (link['source'], link['target'])

    def test_edge_type_is_gateway(self):
        builder = TopologyBuilder()
        result = builder.build_topology(SAMPLE_SCAN)

        for link in result['links']:
            assert link['type'] == 'gateway'

    def test_empty_hosts(self):
        builder = TopologyBuilder()
        result = builder.build_topology({'subnet': '10.0.0.0/24', 'hosts': []})

        assert result['stats']['total_nodes'] == 0
        assert result['links'] == []

    def test_no_gateway_in_hosts(self):
        builder = TopologyBuilder()
        scan = {
            'subnet': '192.168.1.0/24',
            'hosts': [
                {'ip': '192.168.1.50', 'open_ports': [], 'state': 'up'},
                {'ip': '192.168.1.60', 'open_ports': [], 'state': 'up'},
            ],
        }
        result = builder.build_topology(scan)

        assert result['stats']['total_nodes'] == 2
        assert result['links'] == []


class TestGetGraphData:
    def test_output_shape(self):
        builder = TopologyBuilder()
        builder.build_topology(SAMPLE_SCAN)
        data = builder.get_graph_data()

        assert 'nodes' in data
        assert 'links' in data
        assert 'stats' in data
        assert 'total_nodes' in data['stats']
        assert 'total_links' in data['stats']
        assert 'node_types' in data['stats']

    def test_node_fields(self):
        builder = TopologyBuilder()
        builder.build_topology(SAMPLE_SCAN)
        data = builder.get_graph_data()

        node = data['nodes'][0]
        for field in ('id', 'label', 'ip', 'type', 'hostname', 'state',
                      'mac_address', 'vendor', 'open_ports', 'protocols'):
            assert field in node

    def test_node_type_counts(self):
        builder = TopologyBuilder()
        builder.build_topology(SAMPLE_SCAN)
        data = builder.get_graph_data()

        counts = data['stats']['node_types']
        assert counts.get('gateway') == 1
        assert counts.get('server') == 1
        assert counts.get('workstation') == 1
        assert counts.get('printer') == 1



import os
import json
import tempfile
from unittest.mock import patch, MagicMock

from cli import main, _print_table


MOCK_SCAN_RESULTS = {
    'subnet': '192.168.1.0/24',
    'hosts': [
        {
            'ip': '192.168.1.1',
            'hostname': 'gateway',
            'state': 'up',
            'mac_address': 'AA:BB:CC:DD:EE:01',
            'vendor': 'Cisco',
            'open_ports': [],
            'protocols': [],
        },
        {
            'ip': '192.168.1.10',
            'hostname': 'web-server',
            'state': 'up',
            'mac_address': 'AA:BB:CC:DD:EE:10',
            'vendor': 'Dell',
            'open_ports': [{'port': 80, 'protocol': 'tcp', 'service': 'http'}],
            'protocols': ['tcp'],
        },
    ],
    'total_hosts': 2,
    'scan_time': 0,
}


def _mock_scanner():
    """Return a mock NetworkScanner that yields MOCK_SCAN_RESULTS."""
    mock = MagicMock()
    mock.validate_subnet.return_value = True
    mock.get_network_info.return_value = {
        'network_address': '192.168.1.0',
        'broadcast_address': '192.168.1.255',
        'netmask': '255.255.255.0',
        'total_addresses': 256,
        'usable_addresses': 254,
        'prefix_length': 24,
    }
    mock.scan_subnet.return_value = MOCK_SCAN_RESULTS
    return mock


class TestPrintTable:
    def test_prints_hosts(self, capsys):
        _print_table(MOCK_SCAN_RESULTS)
        out = capsys.readouterr().out

        assert '192.168.1.1' in out
        assert '192.168.1.10' in out
        assert 'gateway' in out
        assert 'server' in out

    def test_empty_results(self, capsys):
        _print_table({'hosts': []})
        out = capsys.readouterr().out

        assert 'No hosts found' in out


class TestCLI:
    @patch('cli.NetworkScanner')
    def test_basic_scan(self, mock_cls, capsys):
        mock_cls.return_value = _mock_scanner()
        main(['192.168.1.0/24'])
        out = capsys.readouterr().out

        assert 'Found 2 hosts' in out
        assert '192.168.1.1' in out

    @patch('cli.NetworkScanner')
    def test_invalid_subnet(self, mock_cls):
        mock = MagicMock()
        mock.validate_subnet.return_value = False
        mock_cls.return_value = mock

        try:
            main(['bad-subnet'])
        except SystemExit as e:
            assert e.code == 1

    @patch('cli.NetworkScanner')
    def test_image_svg(self, mock_cls):
        mock_cls.return_value = _mock_scanner()

        with tempfile.NamedTemporaryFile(suffix='.svg', delete=False) as f:
            path = f.name

        try:
            main(['192.168.1.0/24', '-i', path, '-q'])

            with open(path, 'r') as f:
                svg = f.read()
            assert svg.startswith('<svg')
            assert '192.168.1.1' in svg
        finally:
            os.unlink(path)

    @patch('cli.NetworkScanner')
    def test_image_png(self, mock_cls):
        mock_cls.return_value = _mock_scanner()

        with tempfile.NamedTemporaryFile(suffix='.png', delete=False) as f:
            path = f.name

        try:
            main(['192.168.1.0/24', '-i', path, '-q'])

            with open(path, 'rb') as f:
                header = f.read(8)
            assert header[:4] == b'\x89PNG'
        finally:
            os.unlink(path)

    @patch('cli.NetworkScanner')
    def test_image_jpeg(self, mock_cls):
        mock_cls.return_value = _mock_scanner()

        with tempfile.NamedTemporaryFile(suffix='.jpg', delete=False) as f:
            path = f.name

        try:
            main(['192.168.1.0/24', '-i', path, '-q'])

            with open(path, 'rb') as f:
                header = f.read(2)
            assert header == b'\xff\xd8'
        finally:
            os.unlink(path)

    @patch('cli.NetworkScanner')
    def test_json_output(self, mock_cls):
        mock_cls.return_value = _mock_scanner()

        with tempfile.NamedTemporaryFile(suffix='.json', delete=False) as f:
            path = f.name

        try:
            main(['192.168.1.0/24', '--json', path, '-q'])

            with open(path, 'r') as f:
                data = json.load(f)
            assert 'metadata' in data
            assert 'topology' in data
        finally:
            os.unlink(path)

    @patch('cli.NetworkScanner')
    def test_csv_output(self, mock_cls):
        mock_cls.return_value = _mock_scanner()

        with tempfile.NamedTemporaryFile(suffix='.csv', delete=False) as f:
            path = f.name

        try:
            main(['192.168.1.0/24', '--csv', path, '-q'])

            with open(path, 'r') as f:
                content = f.read()
            assert 'IP' in content
            assert '192.168.1.1' in content
        finally:
            os.unlink(path)

    @patch('cli.NetworkScanner')
    def test_quiet_suppresses_table(self, mock_cls, capsys):
        mock_cls.return_value = _mock_scanner()
        main(['192.168.1.0/24', '-q'])
        out = capsys.readouterr().out

        assert '192.168.1.1' not in out

    @patch('cli.NetworkScanner')
    def test_scan_type_flag(self, mock_cls):
        mock = _mock_scanner()
        mock_cls.return_value = mock
        main(['192.168.1.0/24', '-t', 'tcp', '-q'])

        mock.scan_subnet.assert_called_once_with('192.168.1.0/24', 'tcp')



from unittest.mock import MagicMock, patch

from discovery.nmap_scan import NetworkScanner


class TestValidateSubnet:
    def test_valid_cidr(self):
        scanner = NetworkScanner.__new__(NetworkScanner)
        assert scanner.validate_subnet("192.168.1.0/24") is True

    def test_valid_cidr_16(self):
        scanner = NetworkScanner.__new__(NetworkScanner)
        assert scanner.validate_subnet("10.0.0.0/16") is True

    def test_invalid_subnet(self):
        scanner = NetworkScanner.__new__(NetworkScanner)
        assert scanner.validate_subnet("not-a-subnet") is False

    def test_empty_string(self):
        scanner = NetworkScanner.__new__(NetworkScanner)
        assert scanner.validate_subnet("") is False


class TestGetNetworkInfo:
    def test_basic_info(self):
        scanner = NetworkScanner.__new__(NetworkScanner)
        info = scanner.get_network_info("192.168.1.0/24")

        assert info['network_address'] == "192.168.1.0"
        assert info['broadcast_address'] == "192.168.1.255"
        assert info['netmask'] == "255.255.255.0"
        assert info['total_addresses'] == 256
        assert info['usable_addresses'] == 254
        assert info['prefix_length'] == 24

    def test_slash_30(self):
        scanner = NetworkScanner.__new__(NetworkScanner)
        info = scanner.get_network_info("10.0.0.0/30")

        assert info['total_addresses'] == 4
        assert info['usable_addresses'] == 2

    def test_slash_32(self):
        scanner = NetworkScanner.__new__(NetworkScanner)
        info = scanner.get_network_info("10.0.0.1/32")

        assert info['total_addresses'] == 1
        assert info['usable_addresses'] == 0


class TestScanSubnet:
    @patch("discovery.nmap_scan.nmap.PortScanner")
    def test_scan_returns_hosts(self, mock_scanner_cls):
        mock_nm = MagicMock()
        mock_scanner_cls.return_value = mock_nm

        mock_nm.all_hosts.return_value = ["192.168.1.1", "192.168.1.10"]
        mock_nm.__getitem__ = MagicMock()

        host_mock = MagicMock()
        host_mock.state.return_value = "up"
        host_mock.__getitem__ = MagicMock(return_value={})
        host_mock.__contains__ = MagicMock(return_value=False)
        host_mock.all_protocols.return_value = []
        mock_nm.__getitem__.return_value = host_mock

        scanner = NetworkScanner()
        results = scanner.scan_subnet("192.168.1.0/24", "ping")

        assert results['total_hosts'] == 2
        assert results['subnet'] == "192.168.1.0/24"
        assert len(results['hosts']) == 2

    def test_scan_invalid_subnet_raises(self):
        scanner = NetworkScanner.__new__(NetworkScanner)
        scanner.nm = MagicMock()
        scanner.logger = MagicMock()

        try:
            scanner.scan_subnet("invalid", "ping")
            assert False, "Should have raised ValueError"
        except ValueError:
            pass



import ipaddress
import logging
from typing import List, Dict, Any

from config import SERVER_PORTS, PRINTER_PORTS


class TopologyBuilder:
    def __init__(self):
        self.nodes: Dict[str, Dict[str, Any]] = {}
        self.edges: List[Dict[str, Any]] = []
        self.logger = logging.getLogger(__name__)

    def build_topology(self, scan_results: Dict[str, Any]) -> Dict[str, Any]:
        """
        Build network topology from scan results.

        Args:
            scan_results: Results from network scanner

        Returns:
            Dict with nodes and edges representing the topology
        """
        self.nodes.clear()
        self.edges.clear()
        hosts = scan_results.get('hosts', [])
        subnet = scan_results.get('subnet', '')

        if not hosts:
            return self.get_graph_data()

        for host in hosts:
            ip = host['ip']
            self.nodes[ip] = {
                'ip': ip,
                'hostname': host.get('hostname', ''),
                'state': host.get('state', 'unknown'),
                'mac_address': host.get('mac_address', ''),
                'vendor': host.get('vendor', ''),
                'open_ports': host.get('open_ports', []),
                'protocols': host.get('protocols', []),
                'node_type': self._classify_host(host),
            }

        self._infer_connections(hosts, subnet)

        return self.get_graph_data()

    def _classify_host(self, host: Dict[str, Any]) -> str:
        """Classify host type based on IP address and open ports."""
        ip = host['ip']
        if ip.endswith('.1') or ip.endswith('.254'):
            return 'gateway'

        host_ports = {port['port'] for port in host.get('open_ports', [])}

        if host_ports & SERVER_PORTS:
            return 'server'

        if host_ports & PRINTER_PORTS:
            return 'printer'

        return 'workstation'

    def _infer_connections(self, hosts: List[Dict[str, Any]], subnet: str):
        """
        Build a star topology: every host connects to the gateway.

        This is the only topology that can be honestly inferred from a
        scan without SNMP/CDP/LLDP neighbor data.
        """
        try:
            network = ipaddress.ip_network(subnet, strict=False)
            gateway_ip = str(network.network_address + 1)

            actual_gateway = None
            for ip, attrs in self.nodes.items():
                if attrs.get('node_type') == 'gateway':
                    actual_gateway = ip
                    break

            if actual_gateway:
                gateway_ip = actual_gateway

            if gateway_ip not in self.nodes:
                return

            host_ips = [h['ip'] for h in hosts]
            for ip in host_ips:
                if ip != gateway_ip:
                    self.edges.append({
                        'source': ip,
                        'target': gateway_ip,
                        'weight': 1.0,
                        'type': 'gateway',
                    })

        except (ValueError, ipaddress.AddressValueError) as exc:
            self.logger.error("Error inferring connections: %s", exc)

    def get_graph_data(self) -> Dict[str, Any]:
        """Get graph data in a format suitable for D3.js visualization."""
        nodes = []
        for ip, attrs in self.nodes.items():
            nodes.append({
                'id': ip,
                'label': attrs.get('hostname') or ip,
                'ip': ip,
                'type': attrs.get('node_type', 'workstation'),
                'hostname': attrs.get('hostname', ''),
                'state': attrs.get('state', 'unknown'),
                'mac_address': attrs.get('mac_address', ''),
                'vendor': attrs.get('vendor', ''),
                'open_ports': len(attrs.get('open_ports', [])),
                'protocols': attrs.get('protocols', []),
            })

        type_counts: Dict[str, int] = {}
        for attrs in self.nodes.values():
            nt = attrs.get('node_type', 'unknown')
            type_counts[nt] = type_counts.get(nt, 0) + 1

        return {
            'nodes': nodes,
            'links': list(self.edges),
            'stats': {
                'total_nodes': len(nodes),
                'total_links': len(self.edges),
                'node_types': type_counts,
            },
        }
